! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                       3D Morton Curve                        !
!                                                              !
!--------------------------------------------------------------!


!! Simple 3D and 2D Morton curve subroutines.
!!
!! Uses simple bit-interleaving algorithm as described in
!! M.J. Aftosmis, M.J. Berger and S.M. Murman, "Applications of Space-Filling Curves to Cartesian Methods for CFD", 2004
!!
!! Routines accept position and morton-index starting at 1!
!!
!! @author Matthias Lieber
module morton_mod

  use kinds_mod
  implicit none
  private  
  
  public :: morton_c2i, morton2d_c2i, morton2d_i2c
  
contains

!! Get index on Morton curve at given 2D coordinate (coordinates-to-index).
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
pure subroutine morton2d_c2i(lev, p, idx)

  integer(i_k), intent(in)  :: lev    !! level of the morton curve (1=2x2, 2=4x4, 3=8x8, etc)
  integer(i_k), intent(in)  :: p(2)   !! 2D coordinates [1 ... 2^$lev]
  integer(i_k), intent(out) :: idx    !! index on the 2D morton curve [1 ... (2^$lev)^2]
  
  integer(i_k) :: i, mask, shft(2), p0(2)
  
  idx=0
  
  mask=1
  shft(1)=1
  shft(2)=0
  p0=p-1
  
  do i=1,lev
    ! using of ior instead +
    idx  = ior( idx , ior( ishft(iand(p0(1),mask),shft(1)) , ishft(iand(p0(2),mask),shft(2)) ) )
    mask = ishft(mask,1)  ! mask=mask*2
    shft = shft + 1
  end do
  
  idx = idx + 1
  
end subroutine morton2d_c2i


!! Get 2D coordinate at given index on curve (index-to-coordinates).
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
pure subroutine morton2d_i2c(lev, idx, p)

  integer(i_k), intent(in)  :: lev    !! level of the morton curve (1=2x2, 2=4x4, 3=8x8, etc)
  integer(i_k), intent(in)  :: idx    !! index on the 2D morton curve [1 ... (2^$lev)^2]
  integer(i_k), intent(out) :: p(2)   !! 2D coordinates [1 ... 2^$lev]
  
  integer(i_k) :: i, mask(2), shft(2), idx0
  
  mask(1)=2
  mask(2)=1
  shft(1)=-1
  shft(2)=0
  p=0
  idx0=idx-1
  
  do i=1,lev
    p(1) = ior( p(1), ishft( iand(idx0,mask(1)), shft(1) ) )
    p(2) = ior( p(2), ishft( iand(idx0,mask(2)), shft(2) ) )
    shft = shft - 1
    mask = ishft(mask,2)  ! mask=mask*4
  end do
  
  p = p + 1
  
end subroutine morton2d_i2c


!! Get index on Morton curve at given 3D coordinate (coordinates-to-index).
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
pure subroutine morton_c2i(lev, p, idx)

  integer(i_k), intent(in)  :: lev    !! level of the morton curve (1=2x2x2, 2=4x4x4, 3=8x8x8, etc)
  integer(i_k), intent(in)  :: p(3)   !! 3D coordinates [1 ... 2^$lev]
  integer(i_k), intent(out) :: idx    !! index on the 3D morton curve [1 ... (2^$lev)^3]
  
  integer(i_k) :: i, mask, shft(3), p0(3)
  
  idx=0
  
  mask=1
  shft(1)=2
  shft(2)=1
  shft(3)=0
  p0=p-1
  
  do i=1,lev
    ! using of ior instead +
    idx  = ior( ior( idx, ishft(iand(p0(1),mask),shft(1)) )  , & 
           ior( ishft(iand(p0(2),mask),shft(2)) , ishft(iand(p0(3),mask),shft(3)) ) )
    mask = ishft(mask,1)  ! mask=mask*2
    shft = shft + 2
  end do

  idx = idx + 1
  
end subroutine morton_c2i

end module morton_mod
